/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef PARTITION_TREE_HPP
#define PARTITION_TREE_HPP

class Partition_tree {
public:
  int size;                     /* Size of the index set */
  Partition_tree* c[2];         /* Pointers to (2) children */

  Partition_tree() {
    size=0;
    c[0]=NULL;
    c[1]=NULL;
  }

  bool check() {                /* Check that the tree is valid */
    /* Check that the tree is binary (not necessarily complete,
     * but nodes with only 1 child are not acceptable). Also
     * check that the clustering is correct, i.e. that parent
     * index size is the sum of that of children.
     */
    if(size<=0)
      /* Check cluster size */
      return false;
    if(c[0]==NULL && c[1]==NULL)
      /* Leaf */
      return true;
    if(c[0]!=NULL && c[1]==NULL)
      /* Only 1 child: this cannot happen */
      return false;
    if(c[0]==NULL && c[1]!=NULL)
      /* Only 1 child: this cannot happen */
      return false;
    /* Interior node: valid if sizes match and children are valid */
    return (size==c[0]->size+c[1]->size) && c[0]->check() && c[1]->check();
  }

  int treesize() {              /* Compute number of nodes */
    /* Compute the number of nodes in the tree. */
    int s = 1;
    if (c[0]) s += c[0]->treesize();
    if (c[1]) s += c[1]->treesize();
    return s;
  }

  int treeheight() {            /* Compute number of levels */
    /* Compute the number of levels in the tree. */
    int h0 = 0, h1 = 0;
    if (c[0]) h0 = c[0]->treeheight();
    if (c[1]) h1 = c[1]->treeheight();
    return 1+std::max(h0,h1);
  }

  Partition_tree** postorder() {/* List of nodes in postorder */
    /* Return a list of nodes in postorder (left->right) */
    Partition_tree ** porder=new Partition_tree*[treesize()];
    int index=0;
    traverse(porder,&index);
    return porder;
  }

  int* leftchildren() {         /* Indices of left children in postorder */
    /* Return index (in the postorder) of the left child of each node, in postorder */
    int *lc=new int[treesize()];
    int index=0;
    lctraverse(lc,&index);
    return lc;
  }

  int* rightchildren() {        /* Indices of right children in postorder */
    /* Return index (in the postorder) of the right child of each node, in postorder */
    int *rc=new int[treesize()];
    int index=0;
    rctraverse(rc,&index);
    return rc;
  }

  void print() {                /* Print nodes in postorder */
    Partition_tree **porder=postorder();
    int *lc=leftchildren();
    int *rc=rightchildren();
    int n=treesize();
    for(int i=0;i<n;i++) {
      std::cout << "Node " << i << "; size " << porder[i]->size;
      if(porder[i]->c[0]!=NULL)
        std::cout << "; left child " << lc[i] << ": size " << porder[i]->c[0]->size;
      if(porder[i]->c[1]!=NULL)
        std::cout << "; right child " << rc[i] << ": size " << porder[i]->c[1]->size;
      if(porder[i]->c[0]==NULL && porder[i]->c[1]==NULL)
        std::cout << "; leaf.";
      std::cout << std::endl;
    }
  }

private:

  void traverse(Partition_tree **porder, int *index) { /* Auxiliary routine for the postorder traversal */
    /* Traverse left subtree, right subtree, then save node into porder
     * and update the location index.
     */
    if(c[0]!=NULL)
      c[0]->traverse(porder,index);
    if(c[1]!=NULL)
      c[1]->traverse(porder,index);
    porder[(*index)++]=this;
  }

  void lctraverse(int *lc, int *index) {            /* Auxiliary routine for the postorder traversal */
    /* Traverse left subtree, right subtree, then save node into porder
     * and update the location index.
     */
    int tmp;
    if(c[0]!=NULL) {
      c[0]->lctraverse(lc,index);
      tmp=*index-1;
    } else {
      tmp=0;
    }
    if(c[1]!=NULL)
      c[1]->lctraverse(lc,index);
    lc[(*index)++]=tmp;
  }

  void rctraverse(int *rc, int *index) {            /* Auxiliary routine for the postorder traversal */
    /* Traverse left subtree, right subtree, then save node into porder
     * and update the location index.
     */
    int tmp;
    if(c[0]!=NULL)
      c[0]->rctraverse(rc,index);
    if(c[1]!=NULL) {
      c[1]->rctraverse(rc,index);
      tmp=*index-1;
    } else {
      tmp=0;
    }
    rc[(*index)++]=tmp;
  }

};

#endif // PARTITION_TREE_HPP
