/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef SEQ_MGSINTERPOL_HPP
#define SEQ_MGSINTERPOL_HPP

template<typename T, typename S> void seq_MGSInterpol(int m, int n, T *A, S tol, T **X, int *r, int **J, int **P) {
  /* Serial Interpolative Decomposition of A.
   * Decompose the m x n matrix A into A=A(:,J) [I X]*P^T
   * with J a set of columns of size r, and X rk x (n-rk).
   * Step 1: RRQR(A) => A P = Q R = Q [R1 R2]
   * Step 2: TRSM: A P = (Q R1) (R1^-1 R) => A = A(:,J)*[I X]*P^T
   *
   * Matrices are assumed to be stored in column-major format.
   * Matrix A is NOT modified.
   *
   * Matrix Q is initially a copy of A, then it contains the Q factor and is eventually discarded.
   * Matrix X has initial size n x n but is eventually reshaped into r x (b-r) (0-rows and identity columns are discarded).
   *
   * In this version, we perform rank-nbupd updates. The default size is 32; it can
   * be modified with -DMGSBLOCK=x
   *
   * Norms are updated in a fast way instead of being constantly recomputed; with
   * -DMGSNORMS, all norms are computed at each step instead.
   */

  int i, j, k, itmp, minmn, imax;
  int nbupd;
  S tau, firstnorm;
  S *norms;
  T dtmp;
  T *Q, *R;
  bool rank_revealed;

  T NONE=-1;
  T ZERO= 0;
  T  ONE= 1;

  /* Blocksize */
#if defined(MGSBLOCK)
  nbupd=MGSBLOCK;
#else
  if(std::min(m,n)<128)
    /* For small problems, use the BLAS1 version. This helps with accuracy. */
    nbupd=1;
  else
    nbupd=32;
#endif

#if defined(checkID)
  T *B, *C;
  S *work;
#endif

  /* Q: m x n, copy of A */
  Q=new T[m*n];
  for(i=0;i<m*n;i++)
    Q[i]=A[i];

  /* R: n x n */
  R=new T[n*n];
  for(i=0;i<n*n;i++)
    R[i]=ZERO;

  /* J: indices of permuted colums */
  *J=new int[n];
  for(i=0;i<n;i++)
      (*J)[i]=i+1;

  /* P: permutation of columns */
  *P=new int[n];
  for(i=0;i<n;i++)
      (*P)[i]=i+1;

  /* Norms of columns */
  norms=new S[n];
  for(k=0;k<n;k++) {
    norms[k]=std::pow(nrm2<T,S>(m,&Q[k*m],IONE),2);
  }

  /* Step 1: RRQR */
  minmn=std::min(m,n);
  *r=minmn;
  rank_revealed=false;
  i=0;
  j=0;
  while (i<minmn && j<minmn && !rank_revealed) {
    for(i=j;i<=std::min(minmn-1,j+nbupd-1);i++) {
      /* a/ Find the column of A with maximum norm */
      imax=i;
      tau=norms[i];
      for(k=i+1;k<n;k++) {
        if(norms[k]>tau) {
          tau=norms[k];
          imax=k;
        }
      }
      (*P)[i]=imax+1;

      /* b/ Interchange column imax (max norm) and i (current) for both Q and R */
      if(imax>i) {
        for(k=0;k<m;k++) {
          dtmp=Q[imax*m+k];
          Q[imax*m+k]=Q[i*m+k];
          Q[i*m+k]=dtmp;
        }
        for(k=0;k<i;k++) {
          dtmp=R[imax*n+k];
          R[imax*n+k]=R[i*n+k];
          R[i*n+k]=dtmp;
        }
        itmp=(*J)[i];
        (*J)[i]=(*J)[imax];
        (*J)[imax]=itmp;
        norms[imax]=norms[i];
      }

      /* b'/ Update the pivot column with what we have until now in this block, cfr geru */
      gemv('N',m,i-j,NONE,&Q[j*m],m,&R[i*n+j],IONE,ONE,&Q[i*m],IONE);

      /* c/ Compute R(i,i) = ||A(:,i)||
       * Recompute the column norm, this is more accurate than the updated norm in norms[i]
       */
      tau=nrm2<T,S>(m,&Q[i*m],IONE);
      if(i==0) {
        if(tau<tol) {
          *r=0;
          rank_revealed=true;
          break;
        }
        firstnorm=tau;
      }
      if(tau/firstnorm<tol) {
        *r=i;
        rank_revealed=true;
        break;
      }
      R[i*n+i]=static_cast<T>(tau);


      /* d/ Compute Q(:,i) = A(:,i)/R(i,i) */
      scal(m,ONE/static_cast<T>(tau),&Q[i*m],IONE);

      /* e/ Compute R(i,i+1:n) = Q(:,i)'*A(:,i+1:n); computed via transposition: A(:,i+1)'*Q(:,i) */
      gemv('C',m,n-i-1,ONE,&Q[(i+1)*m],m,&Q[i*m],IONE,ZERO,&R[(i+1)*n+i],n);
      lacgv(n-i-1,&R[(i+1)*n+i],n);

      /* e'/ Update norms */
      for(k=i+1;k<n;k++)
        norms[k]-=std::pow(std::abs(R[i+k*n]),2);
    }

    /* f/ Update trailing matrix: columns j:i-1 have been factored.
     * We update the trailing columns rrrrra rank-[(i-1)-(j)+1=i-j] update.
     * Q(:,i:n) -= Q(:,j:i)*R(j:i,i:n).
     * (instead of Q(:,i)*R(i,i:n) in the rank-1 update version).
     */
    gemm('N','N',m,n-i,i-j,NONE,&Q[j*m],m,&R[j+i*n],n,ONE,&Q[i*m],m);

#if defined(MGSNORMS)
    /* In some situations, using exact norms might be more stable. */
    for(k=i;k<n;k++)
      norms[k]=std::pow(nrm2<T,S>(m,&Q[k*m],IONE),2);
#endif

    /* Next block */
    j+=nbupd;
  }

  delete[] norms;

#if defined(debug)
  std::cout << "Rank = " << *r << std::endl << std::endl;
  printMatrix("Q",Q,m,n,m);
  printMatrix("R",R,n,n,n);
#endif

#if defined(checkID)
  /* Copy A into B. Permute B into AP, compute B=AP-QR and norm(B)/norm(A). */
  B=new T[m*n];
  for(i=0;i<m*n;i++)
    B[i]=A[i];
  lapmt(IONE,m,n,B,m,*J);

#if defined(debug)
  printMatrix("AP",B,m,n,m);
#endif

  gemm('N','N',m,n,*r,NONE,Q,m,R,n,ONE,B,m);

#if defined(debug)
  printMatrix("AP-QR",B,m,n,m);
#endif

  work=new S[m];
  std::cout << "Serial RRQR: ||AP-QR||_inf/||A||_inf=      " << lange('I',m,n,B,m,work)/lange('I',m,n,A,m,work) << std::endl << std::endl;
  delete[] B;
  delete[] work;
#endif

  delete[] Q;

  if(*r==0) {
    delete[] R;
    *X=NULL;
    delete[] *J;
    *J=NULL;
    return;
  }

  /* Step 2: TRSM and permutation: R1^-1 R = [I R1^-1 R2] = [I X] with R = [R1 R2], R1 r x r */
  trsm('L','U','N','N',*r,n-*r,ONE,R,n,&R[*r*n],n);

  *X=new T[*r*(n-*r)];
  for(j=0;j<n-*r;j++)
    for(i=0;i<*r;i++)
      (*X)[j*(*r)+i]=R[(*r+j)*n+i];
  delete[] R;

#if defined(checkID)
  /* Copy A into B. Compute C=A(:,J)*[I X] by copying A(:,J) into C(:,1:r) then
   * computing C(:,r+1:n)=C(:,1:r)*X. Then permute into C=C P^T.
   * Finally compute B=B-C=A-A(:,J)*[I X]*P^T and norm(B)/norm(A). */
  B=new T[m*n];
  for(i=0;i<m*n;i++)
    B[i]=A[i];

  C=new T[m*n];
  for(i=0;i<*r;i++)
    lacpy('N',m,1,&A[((*J)[i]-1)*m],m,&C[i*m],m);
  gemm('N','N',m,n-*r,*r,ONE,C,m,*X,*r,ZERO,&C[*r*m],m);
  lapmt(IZERO,m,n,C,m,*J);

  for(i=0;i<m*n;i++)
    B[i]=B[i]-C[i];

  work=new S[m];
  std::cout << "Serial ID: ||A-A(:,J)X||_inf/||A||_inf=    " << lange('I',m,n,B,m,work)/lange('I',m,n,A,m,work) << std::endl << std::endl;
  delete[] B;
  delete[] C;
  delete[] work;
#endif

  int *Jtmp=new int[*r];
  for(int i=0;i<*r;i++)
    Jtmp[i]=(*J)[i];
  delete[] *J;
  *J=Jtmp;

#if defined(debug)
  printMatrix("X",*X,*r,n-*r,*r);
  printMatrix("J",*J,1,*r,1);
  printMatrix("P",*J,1,*r,1);
#endif

}

#endif // SEQ_MGSINTERPOL_HPP
