/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef SEQ_GENHMATRIX_HPP
#define SEQ_GENHMATRIX_HPP

template<typename T, typename S> void seq_genHMatrix(T *A, HSS_tree<T,S> *htree, S perrank) {
  /* Generate a kinda H-matrix where each block has rank perrank times its size.
   * The algorithm is the following:
   *   for each node i in postorder
   *     I=segment associated with i
   *     if i is a leaf
   *       A(I,I)=random matrix of size I + strong diagonal (full rank, and good for factorization)
   *     else
   *       b=length(I)
   *       I{1,2}=segment of child {1,2}
   *       A(I1,I2)= outer product of two random matrices of size b x (perrank*b) (should have rank roughly perrank*b)
   *
   * NB: this is for testing purposes, so it's not a method of HSS_tree.
   */

  int i, j;
  int ib, ie;
  int b, br, bc;
  int r;
  int dim_tree;
  int n;
  T *U, *V, *D;
  HSS_node<T,S> **nodes;
  HSS_node<T,S> *child1, *child2;

  T ZERO=0;
  T  ONE=1;

  n=htree->a_dim;
  dim_tree=htree->dim_tree;
  nodes=htree->nodes;

  StrumpackRand rgen(time(NULL));

  for(i=0;i<dim_tree;i++) {
    ib=nodes[i]->tree_ib-1;
    ie=nodes[i]->tree_ie-1;
    b=ie-ib+1;
    if(nodes[i]->nr_children==0) {
      /* Diagonal blocks are full-rank with large values on the diagonal */
      D=new T[b*b];
      denseRand(D,b,b,&rgen);
      for(j=0;j<b;j++)
        D[j*b+j]=static_cast<T>(n)*static_cast<T>(n);
      lacpy('N',b,b,D,b,&A[ib*n+ib],n);
      delete[] D;
    } else {
      /* Generate the two off-diagonal blocks A(I1,I2) and A(I2,I1) */
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];

      /* A(I1,I2) */
      br=child1->tree_ie-child1->tree_ib+1;
      bc=child2->tree_ie-child2->tree_ib+1;
      /* Rank r: at least one, at most perrank*min(br,bc) */
      r=std::max(1,static_cast<int>(floor(perrank*(std::min(br,bc)))));
      /* U is br*r, V is r*bc */
      U=new T[br*r];
      denseRand(U,br,r,&rgen);
      V=new T[r*bc];
      denseRand(V,r,bc,&rgen);
      gemm('N','N',br,bc,r,ONE,U,br,V,r,ZERO,&A[(child2->tree_ib-1)*n+(child1->tree_ib-1)],n);
      delete[] U;
      delete[] V;

      /* A(I2,I1) */
      br=child2->tree_ie-child2->tree_ib+1;
      bc=child1->tree_ie-child1->tree_ib+1;
      r=std::max(1,static_cast<int>(floor(perrank*(std::min(br,bc)))));
      /* U is br*r, V is r*bc */
      U=new T[br*r];
      denseRand(U,br,r,&rgen);
      V=new T[r*bc];
      denseRand(V,r,bc,&rgen);
      gemm('N','N',br,bc,r,ONE,U,br,V,r,ZERO,&A[(child1->tree_ib-1)*n+(child2->tree_ib-1)],n);
      delete[] U;
      delete[] V;
    }
  }

}

#endif // SEQ_GENHMATRIX_HPP
