/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government igs granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef FRONTAL_MATRIX_DENSE_MPI_HPP
#define FRONTAL_MATRIX_DENSE_MPI_HPP

#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>
#include "CompressedSparseMatrix.hpp"
#include "DistributedMatrix.hpp"
#include "MatrixReordering.hpp"
#include "TaskTimer.hpp"
#include "MPI_wrapper.hpp"
#include "FrontalMatrixMPI.hpp"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t> class FrontalMatrixHSSMPI;

  template<typename scalar_t,typename real_t,typename integer_t>
  class FrontalMatrixDenseMPI : public FrontalMatrixMPI<scalar_t,real_t,integer_t> {
  public:
    FrontalMatrixDenseMPI(CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A,
			  integer_t _sep, integer_t _sep_begin, integer_t _sep_end,
			  integer_t _dim_upd, integer_t* _upd, MPI_Comm _front_comm,
			  int _total_procs);
    FrontalMatrixDenseMPI(const FrontalMatrixDenseMPI&) = delete;
    FrontalMatrixDenseMPI& operator=(FrontalMatrixDenseMPI const&) = delete;
    ~FrontalMatrixDenseMPI();
    void release_work_memory();
    void build_front();
    void partial_factorization();
    void extend_add();
    void extend_add_mpi_fill_buffers(FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>* ch,
				     std::vector<std::vector<scalar_t>>& sbuf);
    void extend_add_mpi_recv_buffers(FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>* ch, scalar_t** pbuf);
    void extend_add_seq_fill_buffers(FrontalMatrixDense<scalar_t,real_t,integer_t>* ch,
				     std::vector<std::vector<scalar_t>>& sbuf);
    void extend_add_seq_recv_buffers(FrontalMatrixDense<scalar_t,real_t,integer_t>* ch, scalar_t** rbuf);

    void UUtxR(integer_t nr_vecs, scalar_t* newR, scalar_t* newSrow, scalar_t* newScol,
	       FrontalMatrix<scalar_t,real_t,integer_t>* parent, int task_depth) {}
    void UUtxR_dist(DistributedMatrix<scalar_t,real_t>& R, DistributedMatrix<scalar_t,real_t>& Srow_upd,
		    DistributedMatrix<scalar_t,real_t>& Scol_upd, int pa_restarts, int task_depth);
    void multifrontal_factorization(HSSStrategy<real_t>& strat, int etree_level=0, int task_depth=0);
    void forward_multifrontal_solve(scalar_t* b_loc, DistributedMatrix<scalar_t,real_t>* b_dist,
				    scalar_t* wmem, int etree_level=0, int task_depth=0);
    void backward_multifrontal_solve(scalar_t* y_loc, DistributedMatrix<scalar_t,real_t>* b_dist,
				     scalar_t* wmem, int etree_level=0, int task_depth=0);
    void get_submatrix(integer_t* rows, integer_t nr_rows, integer_t* cols, integer_t nr_cols,
		       scalar_t* values, integer_t ld, int task_depth) {}
    void get_submatrix_2d(integer_t* rows, integer_t* cols, DistributedMatrix<scalar_t,real_t>& mat, int task_depth);
    long long get_factor_nonzeros(int task_depth=0);
    virtual std::string type() { return "FrontalMatrixDenseMPI"; }
#if defined(dbgHSS)
    friend class FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>;
    void get_F(DistributedMatrix<scalar_t,real_t>& F);
#endif

  private:
    using FDMPI_t = FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>;
    using FD_t = FrontalMatrixDense<scalar_t,real_t,integer_t>;

    DistributedMatrix<scalar_t,real_t> F11;
    DistributedMatrix<scalar_t,real_t> F12;
    DistributedMatrix<scalar_t,real_t> F21;
    DistributedMatrix<scalar_t,real_t> F22;
    int* piv = nullptr;
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::FrontalMatrixDenseMPI
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A, integer_t _sep, integer_t _sep_begin, integer_t _sep_end,
   integer_t _dim_upd, integer_t* _upd, MPI_Comm _front_comm, int _total_procs)
    : FrontalMatrixMPI<scalar_t,real_t,integer_t>(_A, _sep, _sep_begin, _sep_end,
						  _dim_upd, _upd, _front_comm, _total_procs) {}

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::~FrontalMatrixDenseMPI() {
    delete[] piv;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::release_work_memory() {
    F22.clear(); // remove the update block
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::extend_add_mpi_fill_buffers
  (FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>* ch, std::vector<std::vector<scalar_t>>& sbuf) {
    if (ch->F22.active()) {
      std::vector<integer_t> parent_ind;
      integer_t upd_rows_to_sep;
      this->compute_parent_indices(ch, parent_ind, upd_rows_to_sep);
      ch->F22.extend_add_copy_to_buffers(F11, F12, F21, F22, sbuf, this, parent_ind);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::extend_add_mpi_recv_buffers
  (FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>* ch, scalar_t** pbuf) {
    auto F22rank = [&](integer_t r, integer_t c){ return ch->find_rank(r,c,ch->F22); };
    extend_add_copy_from_buffers<scalar_t,real_t,integer_t>
      (F11, F12, F21, F22, pbuf, this->sep_begin, this->upd, ch->upd, ch->dim_upd, F22rank);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::extend_add_seq_fill_buffers
  (FrontalMatrixDense<scalar_t,real_t,integer_t>* ch, std::vector<std::vector<scalar_t>>& sbuf) {
    if (mpi_rank(this->front_comm) == this->child_master(ch)) {
      std::vector<integer_t> pa_ind;
      integer_t to_upd;
      this->compute_parent_indices(ch, pa_ind, to_upd);
      integer_t du = ch->dim_upd;
      integer_t ds = this->dim_sep;
      for (int c=0; c<to_upd; c++) // F11
	for (int r=0; r<to_upd; r++)
	  sbuf[this->find_rank_fixed(pa_ind[r], pa_ind[c], F11)].push_back(ch->F22[r+c*du]);
      for (int c=to_upd; c<du; c++) // F12
	for (int r=0; r<to_upd; r++)
	  sbuf[this->find_rank_fixed(pa_ind[r], pa_ind[c]-ds, F12)].push_back(ch->F22[r+c*du]);
      for (int c=0; c<to_upd; c++) // F21
	for (int r=to_upd; r<du; r++)
	  sbuf[this->find_rank_fixed(pa_ind[r]-ds, pa_ind[c], F21)].push_back(ch->F22[r+c*du]);
      for (int c=to_upd; c<du; c++) // F22
	for (int r=to_upd; r<du; r++)
	  sbuf[this->find_rank_fixed(pa_ind[r]-ds, pa_ind[c]-ds, F22)].push_back(ch->F22[r+c*du]);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::extend_add_seq_recv_buffers
  (FrontalMatrixDense<scalar_t,real_t,integer_t>* ch, scalar_t** pbuf) {
    if (F11.active() || F22.active())
      extend_add_copy_from_buffers<scalar_t,real_t,integer_t>
	(F11, F12, F21, F22, pbuf, this->sep_begin,
	 this->upd, ch->upd, ch->dim_upd, [](integer_t,integer_t){ return 0; });
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::extend_add() {
    if (this->lchild == NULL && this->rchild == NULL) return;
    auto P = mpi_nprocs(this->front_comm);
    std::vector<std::vector<scalar_t>> sbuf(P);
    for (auto ch : {this->lchild, this->rchild}) {
      if (ch && mpi_rank(this->front_comm) == 0)
	{ STRUMPACK_FLOPS(static_cast<long long int>(ch->dim_upd)*ch->dim_upd); }
      if (FDMPI_t* dense_mpi_child = dynamic_cast<FDMPI_t*>(ch))
	extend_add_mpi_fill_buffers(dense_mpi_child, sbuf);
      else if (FD_t* dense_child = dynamic_cast<FD_t*>(ch))
	extend_add_seq_fill_buffers(dense_child, sbuf);
    }
    scalar_t *rbuf=NULL, **pbuf=NULL;
    all_to_all_v(sbuf, rbuf, pbuf, this->front_comm);
    for (auto ch : {this->lchild, this->rchild})
      if (FDMPI_t* dense_mpi_child = dynamic_cast<FDMPI_t*>(ch))
	extend_add_mpi_recv_buffers(dense_mpi_child, pbuf+this->child_master(ch));
      else if (FD_t* dense_child = dynamic_cast<FD_t*>(ch))
	extend_add_seq_recv_buffers(dense_child, pbuf+this->child_master(ch));
    delete[] pbuf;
    delete[] rbuf;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::build_front() {
    if (this->dim_sep) {
      F11 = DistributedMatrix<scalar_t,real_t>(this->ctxt, this->dim_sep, this->dim_sep);
      F11.extract_F11(this->A, this->sep_begin, this->dim_sep);
      if (this->dim_upd) {
	F12 = DistributedMatrix<scalar_t,real_t>(this->ctxt, this->dim_sep, this->dim_upd);
	F12.extract_F12(this->A, this->sep_begin, this->sep_end, this->dim_upd, this->upd);
	F21 = DistributedMatrix<scalar_t,real_t>(this->ctxt, this->dim_upd, this->dim_sep);
	F21.extract_F21(this->A, this->sep_end, this->sep_begin, this->dim_upd, this->upd);
      }
    }
    if (this->dim_upd) {
      F22 = DistributedMatrix<scalar_t,real_t>(this->ctxt, this->dim_upd, this->dim_upd);
      F22.zero();
    }
    extend_add();
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::partial_factorization() {
    if (this->dim_sep && F11.active()) {
#if defined(WRITE_ROOT)
      if (etree_level == 0) {
	if (!mpi_rank(this->front_comm)) std::cout << "Writing root node to file..." << std::endl;
	F11.MPI_binary_write();
	if (!mpi_rank(this->front_comm)) std::cout << "Done. Early abort." << std::endl;
	MPI_Finalize();
	exit(0);
      }
#endif
      piv = new int[F11.local_rows() + F11.MB()];
      F11.LU(piv);
      if (this->dim_upd) {
	F12.lapiv(piv, F11);
	trsm('L', 'L', 'N', 'U', scalar_t(1.), F11, F12);
	trsm('R', 'U', 'N', 'N', scalar_t(1.), F11, F21);
	gemm('N', 'N', scalar_t(-1.), F21, F12, scalar_t(1.), F22);
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::multifrontal_factorization
  (HSSStrategy<real_t>& strat, int etree_level, int task_depth) {
    if (this->visit(this->lchild)) this->lchild->multifrontal_factorization(strat, etree_level+1, task_depth);
    if (this->visit(this->rchild)) this->rchild->multifrontal_factorization(strat, etree_level+1, task_depth);
    build_front();
    if (this->lchild) this->lchild->release_work_memory();
    if (this->rchild) this->rchild->release_work_memory();
    partial_factorization();
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::forward_multifrontal_solve
  (scalar_t* b_loc, DistributedMatrix<scalar_t,real_t>* b_dist, scalar_t* wmem, int etree_level, int task_depth) {
    if (this->visit(this->lchild)) this->lchild->forward_multifrontal_solve(b_loc, b_dist, wmem, etree_level, task_depth);
    if (this->visit(this->rchild)) this->rchild->forward_multifrontal_solve(b_loc, b_dist, wmem, etree_level, task_depth);
    DistributedMatrixWrapper<scalar_t,real_t> Bupd(this->ctxt, this->dim_upd, 1, wmem+this->p_wmem);
    Bupd.zero();
    this->look_left(b_dist[this->sep], wmem);
    if (this->dim_sep) {
      TIMER_TIME(SOLVE_LOWER, 0, t_s);
      b_dist[this->sep].lapiv(piv, F11);
      trsv('L', 'N', 'U', F11, b_dist[this->sep]);
      if (this->dim_upd)
	gemv('N', scalar_t(-1.), F21, b_dist[this->sep], scalar_t(1.), Bupd);
      TIMER_STOP(t_s);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::backward_multifrontal_solve
  (scalar_t* y_loc, DistributedMatrix<scalar_t,real_t>* y_dist, scalar_t* wmem, int etree_level, int task_depth) {
    if (this->dim_sep) {
      TIMER_TIME(SOLVE_UPPER, 0, t_s);
      if (this->dim_upd) {
	DistributedMatrixWrapper<scalar_t,real_t> Yupd(this->ctxt, this->dim_upd, 1, wmem+this->p_wmem);
	gemv('N', scalar_t(-1.), F12, Yupd, scalar_t(1.), y_dist[this->sep]);
      }
      trsv('U', 'N', 'N', F11, y_dist[this->sep]);
      TIMER_STOP(t_s);
    }
    this->look_right(y_dist[this->sep], wmem);
    if (this->visit(this->lchild)) this->lchild->backward_multifrontal_solve(y_loc, y_dist, wmem, etree_level, task_depth);
    if (this->visit(this->rchild)) this->rchild->backward_multifrontal_solve(y_loc, y_dist, wmem, etree_level, task_depth);
  }

  /**
   * Note that mat should be defined on the same context as used in
   * this front. This simplifies communication.
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::get_submatrix_2d
  (integer_t* rows, integer_t* cols, DistributedMatrix<scalar_t,real_t>& mat, int task_depth) {
    if (this->front_comm == MPI_COMM_NULL) return;
    std::vector<integer_t> my_cols, orig_cols, my_rows, orig_rows;
    this->find_my_upd_indices(mat.cols(), cols, my_cols, orig_cols);
    this->find_my_upd_indices(mat.rows(), rows, my_rows, orig_rows);
    auto P = mpi_nprocs(this->front_comm);
    std::vector<std::vector<scalar_t>> sbuf(P);
    F22.extract_copy_to_buffers(my_rows, my_cols, orig_rows, orig_cols, mat, sbuf);
    scalar_t *rbuf=NULL, **pbuf=NULL;
    all_to_all_v(sbuf, rbuf, pbuf, this->front_comm);
    mat.extract_copy_from_buffers(my_rows, my_cols, orig_rows, orig_cols, F22, pbuf);
    delete[] rbuf;
    delete[] pbuf;
  }

  /**
   * build random matrix R_upd of size dim_upd x nr_random_vecs
   * compute
   *   Srow_upd = F22 * R_upd
   *   Scol_upd = F22^* * R_upd
   * do skinny_extend_add of Srow_upd and Scol_upd to the HSSMPI parent front
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::UUtxR_dist
  (DistributedMatrix<scalar_t,real_t>& R, DistributedMatrix<scalar_t,real_t>& Srow_upd,
   DistributedMatrix<scalar_t,real_t>& Scol_upd, int pa_restarts, int task_depth) {
    if (F11.active() || F22.active()) {
      auto b = R.cols();
      DistributedMatrix<scalar_t,real_t> R_upd(this->ctxt, this->dim_upd, b);
      std::function<integer_t(integer_t)> global_row = [&](integer_t r) -> integer_t { return this->upd[r]; };
      R_upd.fill_random(global_row, pa_restarts);
      Srow_upd = DistributedMatrix<scalar_t,real_t>(this->ctxt, this->dim_upd, b);
      Scol_upd = DistributedMatrix<scalar_t,real_t>(this->ctxt, this->dim_upd, b);
      gemm('N', 'N', scalar_t(1.), F22, R_upd, scalar_t(0.), Srow_upd);
      gemm('C', 'N', scalar_t(1.), F22, R_upd, scalar_t(0.), Scol_upd);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> long long
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::get_factor_nonzeros(int task_depth) {
    return this->get_dense_factor_nonzeros(task_depth);
  }


#if defined(dbgHSS)
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>::get_F
  (DistributedMatrix<scalar_t,real_t>& F) {
    F = DistributedMatrix<scalar_t,real_t>(this->ctxt, this->dim_blk, this->dim_blk);
    pgemr2d(this->dim_sep, this->dim_sep, F11.data(), 1, 1, F11.desc(), F.data(), 1, 1, F.desc(), this->ctxt_all);
    pgemr2d(this->dim_sep, this->dim_upd, F12.data(), 1, 1, F12.desc(), F.data(), 1, this->dim_sep+1, F.desc(), this->ctxt_all);
    pgemr2d(this->dim_upd, this->dim_sep, F21.data(), 1, 1, F21.desc(), F.data(), this->dim_sep+1, 1, F.desc(), this->ctxt_all);
    pgemr2d(this->dim_upd, this->dim_upd, F22.data(), 1, 1, F22.desc(), F.data(), this->dim_sep+1, this->dim_sep+1, F.desc(), this->ctxt_all);
  }
#endif

} // end namespace strumpack

#endif //FRONTAL_MATRIX_DENSE_MPI_HPP
