/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government igs granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef FRONTAL_MATRIX_HSS_MPI_HPP
#define FRONTAL_MATRIX_HSS_MPI_HPP

#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>
#include "CompressedSparseMatrix.hpp"
#include "DistributedMatrix.hpp"
#include "MatrixReordering.hpp"
#include "TaskTimer.hpp"
#include "MPI_wrapper.hpp"
#include "FrontalMatrixMPI.hpp"

#define DOSYNC_ALLGATHER_IJ
#define ELEMENTS_BARRIER void();
#include "DenseMPI/StrumpackDensePackage.hpp"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t>
  class FrontalMatrixHSSMPI : public FrontalMatrixMPI<scalar_t,real_t,integer_t> {
  public:
    FrontalMatrixHSSMPI(CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A,
			integer_t _sep, integer_t _sep_begin, integer_t _sep_end,
			integer_t _dim_upd, integer_t* _upd, MPI_Comm _front_comm,
			int _total_procs);
    FrontalMatrixHSSMPI(const FrontalMatrixHSSMPI&) = delete;
    FrontalMatrixHSSMPI& operator=(FrontalMatrixHSSMPI const&) = delete;
    ~FrontalMatrixHSSMPI();
    void release_work_memory();
    void set_HSS_partitioning(HSSPartitionTree& sep_tree, bool is_root);
    void extend_add(int task_depth) {}
    void random_sampling(scalar_t* Rr, scalar_t* Rc, int* descR, scalar_t* Sr, scalar_t* Sc, int* descS);
    void multiply_random(int task_depth);
    void UUtxR(integer_t cols, scalar_t* newR, scalar_t* newSrow, scalar_t* newScol,
	       FrontalMatrix<scalar_t,real_t,integer_t>* pa, int task_depth) {}
    void children_UUtxR(int task_depth);
    void child_UUtxR_seqseq(int task_depth);
    void UUtxR_dist(DistributedMatrix<scalar_t,real_t>& R, DistributedMatrix<scalar_t,real_t>& Srow_upd,
		    DistributedMatrix<scalar_t,real_t>& Scol_upd, int pa_restarts, int task_depth);
    void skinny_extend_add(DistributedMatrix<scalar_t,real_t>& Sr_upd_l, DistributedMatrix<scalar_t,real_t>& Sc_upd_l,
			   DistributedMatrix<scalar_t,real_t>& Sr_upd_r, DistributedMatrix<scalar_t,real_t>& Sc_upd_r);
    void multifrontal_factorization(HSSStrategy<real_t>& strat, int etree_level=0, int task_depth=0);
    void forward_multifrontal_solve(scalar_t* b_loc, DistributedMatrix<scalar_t,real_t>* b_dist,
				    scalar_t* wmem, int etree_level=0, int task_depth=0);
    void backward_multifrontal_solve(scalar_t* y_loc, DistributedMatrix<scalar_t,real_t>* b_dist,
				     scalar_t* wmem, int etree_level=0, int task_depth=0);
    void get_submatrix(integer_t* rows, integer_t nr_rows, integer_t* cols, integer_t nr_cols,
		       scalar_t* values, integer_t ld, int task_depth) {}
    void get_submatrix_2d(integer_t* rows, integer_t* cols, DistributedMatrix<scalar_t,real_t>& mat, int task_depth);
    virtual long long get_node_factor_nonzeros();
    bool isHSS() { return true; };
    std::string type() { return "FrontalMatrixHSSMPI"; }
    integer_t get_maximum_rank(int task_depth) const;

    using FMPI_t = FrontalMatrixMPI<scalar_t,real_t,integer_t>;

    int restarts;
    StrumpackDensePackage<scalar_t,real_t>* hss_mat;

    DistributedMatrixWrapper<scalar_t,real_t> Rrow;
    DistributedMatrixWrapper<scalar_t,real_t> Rcol;
    DistributedMatrixWrapper<scalar_t,real_t> Srow;
    DistributedMatrixWrapper<scalar_t,real_t> Scol;
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::FrontalMatrixHSSMPI
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A, integer_t _sep, integer_t _sep_begin,
   integer_t _sep_end, integer_t _dim_upd, integer_t* _upd, MPI_Comm _front_comm, int _total_procs)
    : FrontalMatrixMPI<scalar_t,real_t,integer_t>
    (_A, _sep, _sep_begin, _sep_end, _dim_upd, _upd, _front_comm, _total_procs), restarts(0) {
    if (this->front_comm != MPI_COMM_NULL && this->dim_blk) {
      hss_mat = new StrumpackDensePackage<scalar_t,real_t>(this->front_comm);
      hss_mat->use_HSS = true;
      hss_mat->verbose = false;
      if (this->dim_sep && this->dim_upd)
	hss_mat->split_HSS = this->dim_sep;
    } else hss_mat = NULL;
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::~FrontalMatrixHSSMPI() {
   if (hss_mat) {
     destroy_Partition_tree(hss_mat->utree);
     hss_mat->utree = NULL;
     delete hss_mat;
   }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::release_work_memory() {
    if (hss_mat) hss_mat->destroy_A22();
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::set_HSS_partitioning(HSSPartitionTree& sep_tree, bool is_root) {
    if (hss_mat) {
      assert(sep_tree.size == this->dim_sep);
      if (this->dim_upd == 0) {
    	if (this->dim_sep == 0) return;
    	else hss_mat->utree = HSSPartitionTree_to_Partition_tree(sep_tree);
      } else {
    	HSSPartitionTree cb_tree(this->dim_upd);
    	cb_tree.refine(params::hss_minimum_partition_size);
    	if (this->dim_sep == 0) hss_mat->utree = HSSPartitionTree_to_Partition_tree(cb_tree);
    	else {
    	  HSSPartitionTree f_tree(this->dim_blk);
    	  f_tree.c.push_back(sep_tree);
    	  f_tree.c.push_back(cb_tree);
    	  hss_mat->utree = HSSPartitionTree_to_Partition_tree(f_tree);
    	}
      }
    }
  }

  /**
   * Simply generate Rrow as a random matrix, but in such a way that
   * rows which also occur in the sibling or in the parent are the same
   * there.
   * Rcol is just a copy of Rrow.
   * Compute F * Rrow and F^T Rcol.
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::random_sampling
  (scalar_t* Rr, scalar_t* Rc, int* descR, scalar_t* Sr, scalar_t* Sc, int* descS) {
    int d = descR[BLACSn];        assert(descR[BLACSrsrc] == 0 && descR[BLACScsrc] == 0);
    int M = descR[BLACSm];	  assert(M == this->dim_blk);
    int MB = descR[BLACSmb];	  assert((MB == DistributedMatrix<scalar_t,real_t>::default_MB));
    int NB = descR[BLACSnb];	  assert((NB == DistributedMatrix<scalar_t,real_t>::default_NB));
    int ctxt = descR[BLACSctxt];
    Rrow = DistributedMatrixWrapper<scalar_t,real_t>(ctxt, M, d, MB, NB, Rr);
    Rcol = DistributedMatrixWrapper<scalar_t,real_t>(ctxt, M, d, MB, NB, Rc);
    Srow = DistributedMatrixWrapper<scalar_t,real_t>(ctxt, M, d, MB, NB, Sr);
    Scol = DistributedMatrixWrapper<scalar_t,real_t>(ctxt, M, d, MB, NB, Sc);
    TIMER_TIME(RANDOM_GENERATE, 1, t_gen);
    std::function<integer_t(integer_t)> global_row = [&](integer_t r) -> integer_t {
      return (r < this->dim_sep) ? r + this->sep_begin : this->upd[r-this->dim_sep];
    };
    Rrow.fill_random(global_row, restarts);
    Rcol.copy(Rrow);
    TIMER_STOP(t_gen);
    TIMER_TIME(RANDOM_SAMPLING, 0, t_sampling);
    multiply_random(0);
    TIMER_STOP(t_sampling);
    restarts++;
  }

  /**
   * Compute
   *   Srow = F Rrow
   *   Scol = F^* Rcol
   * where the front
   *   F = A(sep,sep) + U_1 + U_2
   * A(sep,sep) is part of the sparse matrix.
   * U_1 is the contribution block of child 1.
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::multiply_random(int task_depth) {
    Srow.zero();
    Scol.zero();
    TIMER_TIME(FRONT_MULTIPLY_2D, 1, t_fmult);
    this->A->front_multiply_2d(this->sep_begin, this->sep_end, this->upd, this->dim_upd,
			       Rrow, Srow, Scol, this->ctxt_all, this->front_comm, task_depth);
    TIMER_STOP(t_fmult);
    TIMER_TIME(UUTXR, 1, t_UUtxR);
    children_UUtxR(task_depth);
    TIMER_STOP(t_UUtxR);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::children_UUtxR(int task_depth) {
    if (!this->lchild && !this->rchild) return;
    // both children are sequential
    if (this->lchild && !this->lchild->isMPI() && this->rchild && !this->rchild->isMPI()) {
      child_UUtxR_seqseq(task_depth);
      return;
    }
    // both children are MPI
    if (this->lchild->isMPI() && this->rchild->isMPI()) {
      auto lch = dynamic_cast<FMPI_t*>(this->lchild);
      auto rch = dynamic_cast<FMPI_t*>(this->rchild);
      DistributedMatrix<scalar_t,real_t> Sr_upd_l, Sc_upd_l, Sr_upd_r, Sc_upd_r;
      lch->UUtxR_dist(Rrow, Sr_upd_l, Sc_upd_l, restarts, task_depth);
      rch->UUtxR_dist(Rrow, Sr_upd_r, Sc_upd_r, restarts, task_depth);
      TIMER_TIME(SKINNY_EXTEND_ADD_MPIMPI, 2, t_sea);
      skinny_extend_add(Sr_upd_l, Sc_upd_l, Sr_upd_r, Sc_upd_r);
      TIMER_STOP(t_sea);
      return;
    }
    // one child is seq or NULL, the other is MPI or NULL
    FrontalMatrix<scalar_t,real_t,integer_t>* ch_seq=NULL;
    FMPI_t* ch_mpi = dynamic_cast<FMPI_t*>(this->lchild);
    if (!ch_mpi) ch_mpi = dynamic_cast<FMPI_t*>(this->rchild);
    if (ch_mpi) ch_seq = (ch_mpi == this->lchild) ? this->rchild : this->lchild;
    else ch_seq = this->lchild ? this->lchild : this->rchild;
    scalar_t *Rseq=NULL, *Srowseq=NULL, *Scolseq=NULL;
    int m, n, desc[9];
    if (ch_seq) {
      m = Rrow.rows(), n = Rrow.cols();
      auto N = m*n;
      auto pch = this->child_master(ch_seq);
      auto p = mpi_rank(this->front_comm);
      if (p == pch) {
	Rseq = new scalar_t[3*N];
	Srowseq = Rseq + N;
	Scolseq = Srowseq + N;
	std::fill(Srowseq, Srowseq+2*N, scalar_t(.0));
      }
      /* this->ctxt_all is a row of processes, so set the coordinate just along the row [0, pch] */
      descset(desc, m, n, m, n, 0, pch, this->ctxt_all, m);
      pgemr2d(m, n, Rrow.data(), 1, 1, Rrow.desc(), Rseq, 1, 1, desc, this->ctxt_all);
      if (p == pch) ch_seq->UUtxR(n, Rseq, Srowseq, Scolseq, this, task_depth);
    }
    if (ch_mpi) {
      DistributedMatrix<scalar_t,real_t> Sr_upd, Sc_upd, S_dummy;
      ch_mpi->UUtxR_dist(Rrow, Sr_upd, Sc_upd, restarts, task_depth);
      TIMER_TIME(SKINNY_EXTEND_ADD_MPI1, 2, t_sea);
      if (ch_mpi == this->lchild) skinny_extend_add(Sr_upd, Sc_upd, S_dummy, S_dummy);
      else skinny_extend_add(S_dummy, S_dummy, Sr_upd, Sc_upd);
      TIMER_STOP(t_sea);
    }
    if (ch_seq) {
      TIMER_TIME(SKINNY_EXTEND_ADD_SEQ1, 2, t_sea);
      DistributedMatrix<scalar_t,real_t> Stmp(this->ctxt, m, n);
      pgemr2d(m, n, Srowseq, 1, 1, desc, Stmp.data(), 1, 1, Stmp.desc(), this->ctxt_all);
      geadd('N', scalar_t(1.), Stmp, scalar_t(1.), Srow);
      pgemr2d(m, n, Scolseq, 1, 1, desc, Stmp.data(), 1, 1, Stmp.desc(), this->ctxt_all);
      geadd('N', scalar_t(1.), Stmp, scalar_t(1.), Scol);
      TIMER_STOP(t_sea);
      delete[] Rseq;
    }
  }

  /**
   * Both children are sequential. This means this front only has 2
   * processes working on it. Both these processes are the master of 1
   * of the children, so they will both need to construct a sequential
   * copy of R, then compute Srowseq and Scolseq, and add it to the 2D
   * block-cyclicly distributed Srow and Scol of the parent.
   *
   * It is very important to combine this function for the two
   * children together, to avoid that the children are handled one
   * after the other!!
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::child_UUtxR_seqseq(int task_depth) {
    auto m = Rrow.rows();
    auto n = Rrow.cols();
    auto N = m*n;
    auto p = mpi_rank(this->front_comm);
    auto pl = this->child_master(this->lchild);
    auto pr = this->child_master(this->rchild);
    auto Rseq = new scalar_t[3*N];
    auto Srowseq = Rseq + N;
    auto Scolseq = Srowseq + N;
    std::fill(Srowseq, Srowseq+2*N, scalar_t(.0));
    int dRl[9], dRr[9];
    descset(dRl, m, n, m, n, 0, pl, this->ctxt_all, m);
    descset(dRr, m, n, m, n, 0, pr, this->ctxt_all, m);
    pgemr2d(m, n, Rrow.data(), 1, 1, Rrow.desc(), (p==pl) ? Rseq : NULL, 1, 1, dRl, this->ctxt_all);
    pgemr2d(m, n, Rrow.data(), 1, 1, Rrow.desc(), (p==pr) ? Rseq : NULL, 1, 1, dRr, this->ctxt_all);
    if (p == pl) this->lchild->UUtxR(n, Rseq, Srowseq, Scolseq, this, task_depth);
    if (p == pr) this->rchild->UUtxR(n, Rseq, Srowseq, Scolseq, this, task_depth);
    TIMER_TIME(SKINNY_EXTEND_ADD_SEQSEQ, 2, t_sea);
    DistributedMatrix<scalar_t,real_t> Stmp(this->ctxt, m, n);
    pgemr2d(m, n, (p==pl) ? Srowseq : NULL, 1, 1, dRl, Stmp.data(), 1, 1, Stmp.desc(), this->ctxt_all);
    geadd('N', scalar_t(1.), Stmp, scalar_t(1.), Srow);
    pgemr2d(m, n, (p==pl) ? Scolseq : NULL, 1, 1, dRl, Stmp.data(), 1, 1, Stmp.desc(), this->ctxt_all);
    geadd('N', scalar_t(1.), Stmp, scalar_t(1.), Scol);
    pgemr2d(m, n, (p==pr) ? Srowseq : NULL, 1, 1, dRr, Stmp.data(), 1, 1, Stmp.desc(), this->ctxt_all);
    geadd('N', scalar_t(1.), Stmp, scalar_t(1.), Srow);
    pgemr2d(m, n, (p==pr) ? Scolseq : NULL, 1, 1, dRr, Stmp.data(), 1, 1, Stmp.desc(), this->ctxt_all);
    geadd('N', scalar_t(1.), Stmp, scalar_t(1.), Scol);
    TIMER_STOP(t_sea);
    delete[] Rseq;
  }

  /**
   * compute
   *   Srow_upd = (HSS_22 - theta Vhat^ phi) * Rrow[dim_upd:dim_blk-1,:]
   *   Scol_upd = (HSS_22 - theta Vhat^ phi)^* * Rcol[dim_upd:dim_blk-1,:]
   * and do skinny-extend-add to the parent
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::UUtxR_dist
  (DistributedMatrix<scalar_t,real_t>& R, DistributedMatrix<scalar_t,real_t>& Srow_upd,
   DistributedMatrix<scalar_t,real_t>& Scol_upd, int pa_restarts, int task_depth) {
    if (!this->dim_upd) return;
    if (this->front_comm != MPI_COMM_NULL) {
      auto b = R.cols();
      DistributedMatrix<scalar_t,real_t> R_upd(this->ctxt, this->dim_upd, b);
      std::function<integer_t(integer_t)> global_row = [&](integer_t r) -> integer_t { return this->upd[r]; };
      R_upd.fill_random(global_row, pa_restarts);
      Srow_upd = DistributedMatrix<scalar_t,real_t>(this->ctxt, this->dim_upd, b);
      Scol_upd = DistributedMatrix<scalar_t,real_t>(this->ctxt, this->dim_upd, b);
      if (this->dim_sep) {
	MPI_Pcontrol(1, "DenseMPI::schur_product");
	TIMER_TIME(HSS_SCHUR_PRODUCT, 2, t_sprod);
	hss_mat->schur_product('N', scalar_t(1.), R_upd.data(), R_upd.desc(),
			       scalar_t(0.), Srow_upd.data(), Srow_upd.desc());
	hss_mat->schur_product('C', scalar_t(1.), R_upd.data(), R_upd.desc(),
			       scalar_t(0.), Scol_upd.data(), Scol_upd.desc());
	TIMER_STOP(t_sprod);
	MPI_Pcontrol(-1, "DenseMPI::schur_product");
      } else {
	int descA[9];
	descset(descA, this->dim_blk, this->dim_blk, R_upd.MB(), R_upd.NB(),
		0, 0, this->blacs_context(), this->dim_blk);
	MPI_Pcontrol(1, "DenseMPI::product");
	TIMER_TIME(HSS_PRODUCT, 2, t_prod);
	hss_mat->product('N', scalar_t(1.), NULL, descA, R_upd.data(), R_upd.desc(),
			 scalar_t(0.), Srow_upd.data(), Srow_upd.desc());
	hss_mat->product('C', scalar_t(1.), NULL, descA, R_upd.data(), R_upd.desc(),
			 scalar_t(0.), Scol_upd.data(), Scol_upd.desc());
	TIMER_STOP(t_prod);
	MPI_Pcontrol(-1, "DenseMPI::product");
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::skinny_extend_add
  (DistributedMatrix<scalar_t,real_t>& Sr_upd_l, DistributedMatrix<scalar_t,real_t>& Sc_upd_l,
   DistributedMatrix<scalar_t,real_t>& Sr_upd_r, DistributedMatrix<scalar_t,real_t>& Sc_upd_r) {
    auto P = mpi_nprocs(this->front_comm);
    std::vector<std::vector<scalar_t>> sbuf(P);
    auto lch = dynamic_cast<FMPI_t*>(this->lchild);
    auto rch = dynamic_cast<FMPI_t*>(this->rchild);
    if (Sr_upd_l.active()) {
      std::vector<integer_t> parent_ind;
      integer_t upd_rows_to_sep;
      this->compute_parent_indices(lch, parent_ind, upd_rows_to_sep);
      Sr_upd_l.skinny_extend_add_copy_to_buffers(Srow, sbuf, this, parent_ind);
      Sc_upd_l.skinny_extend_add_copy_to_buffers(Scol, sbuf, this, parent_ind);
    }
    if (Sr_upd_r.active()) {
      std::vector<integer_t> parent_ind;
      integer_t upd_rows_to_sep;
      this->compute_parent_indices(rch, parent_ind, upd_rows_to_sep);
      Sr_upd_r.skinny_extend_add_copy_to_buffers(Srow, sbuf, this, parent_ind);
      Sc_upd_r.skinny_extend_add_copy_to_buffers(Scol, sbuf, this, parent_ind);
    }

    scalar_t *rbuf=NULL, **pbuf=NULL;
    all_to_all_v(sbuf, rbuf, pbuf, this->front_comm);

    if (lch) { // unpack left child contribution
      std::function<int(integer_t,integer_t)> Sl_upd_rank = [&](integer_t r, integer_t c){
	return lch->find_rank(r, c, Sr_upd_l); };
      Srow.skinny_extend_add_copy_from_buffers(pbuf, this->sep_begin, this->dim_sep, this->upd, lch->upd, lch->dim_upd, Sl_upd_rank);
      Scol.skinny_extend_add_copy_from_buffers(pbuf, this->sep_begin, this->dim_sep, this->upd, lch->upd, lch->dim_upd, Sl_upd_rank);
    }
    if (rch) { // unpack right child contribution
      std::function<int(integer_t,integer_t)> Sr_upd_rank = [&](integer_t r, integer_t c){
	return this->child_master(rch) + rch->find_rank(r, c, Sr_upd_r); };
      Srow.skinny_extend_add_copy_from_buffers(pbuf, this->sep_begin, this->dim_sep, this->upd, rch->upd, rch->dim_upd, Sr_upd_rank);
      Scol.skinny_extend_add_copy_from_buffers(pbuf, this->sep_begin, this->dim_sep, this->upd, rch->upd, rch->dim_upd, Sr_upd_rank);
    }
    delete[] pbuf;
    delete[] rbuf;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::multifrontal_factorization
  (HSSStrategy<real_t>& strat, int etree_level, int task_depth) {
    if (this->visit(this->lchild)) this->lchild->multifrontal_factorization(strat, etree_level+1, task_depth);
    if (this->visit(this->rchild)) this->rchild->multifrontal_factorization(strat, etree_level+1, task_depth);
    if (!this->dim_blk) return;

    hss_mat->obj = static_cast<void*>(this); // data argument in element_extraction
    auto element_extraction = [](void *data, int *I, int *J, scalar_t *val, int *desc) {
      // the row and column indices in I and J are local to the front, ie, in [0:dim_sep+dim_upd]
      // in order to extract from the sparse matrix and from the contribution blocks of the children,
      // this needs to be transformed to global indices, in [sep_begin:sep_end+dim_upd]
      auto nr_rows = desc[BLACSm];
      auto nr_cols = desc[BLACSn];
      if (!nr_rows || !nr_cols) return;
      auto f = static_cast<FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>*>(data);
      auto rows_global = new integer_t[nr_rows+nr_cols];
      auto cols_global = rows_global + nr_rows;
      for (integer_t r=0; r<nr_rows; r++)
	rows_global[r] = (I[r]-1 < f->dim_sep) ? I[r]-1+f->sep_begin : f->upd[I[r]-1-f->dim_sep];
      for (integer_t c=0; c<nr_cols; c++)
	cols_global[c] = (J[c]-1 < f->dim_sep) ? J[c]-1+f->sep_begin : f->upd[J[c]-1-f->dim_sep];
      // extract_2d (defined in FrontalMatrix) extracts from sparse
      // matrix and from children's contribution block
      DistributedMatrixWrapper<scalar_t,real_t> mat(desc[BLACSctxt], nr_rows, nr_cols, val);
      TIMER_TIME(EXTRACT_2D, 1, t_ex);
      f->extract_2d(rows_global, cols_global, mat, 0);
      TIMER_STOP(t_ex);
      delete[] rows_global;
    };

    auto Amult = [](void *data, scalar_t *Rr, scalar_t *Rc, int *descR,
		    scalar_t *Sr, scalar_t *Sc, int *descS) {
      static_cast<FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>*>(data)->random_sampling
      (Rr, Rc, descR, Sr, Sc, descS);
    };

    hss_mat->min_rand_HSS = params::random_vector_blocksize;
    hss_mat->max_rand_HSS = params::maximum_rank;
    hss_mat->lim_rand_HSS = params::rank_offset;
    hss_mat->inc_rand_HSS = params::random_vector_blocksize;
    hss_mat->tol_HSS = strat.rctol;
    hss_mat->ctxtglob = this->ctxt_all;
    MPI_Pcontrol(1, "DenseMPI::compress");
    TIMER_TIME(HSS_COMPRESS, 0, t_compress);
    int descA[9];
    descset(descA, this->dim_blk, this->dim_blk, DistributedMatrix<scalar_t,real_t>::default_MB,
	    DistributedMatrix<scalar_t,real_t>::default_NB, 0, 0, this->ctxt, this->dim_blk);
    hss_mat->compress(element_extraction, Amult, descA);
    TIMER_STOP(t_compress);
    MPI_Pcontrol(-1, "DenseMPI::compress");

    if (params::log_ranks) {
#if defined(dbgHSS)
      real_t acc_total = (hss_mat) ? hss_mat->check_compression(fe.data(), fe.desc()) : 0.0;
#else
      real_t acc_total = -1.0;
#endif
      DistributedMatrix<scalar_t,real_t> e(this->ctxt, this->dim_blk, this->dim_blk);
      auto I = new int[2*this->dim_blk];
      auto J = I+this->dim_blk;
      for (int i=0; i<this->dim_blk; i++) I[i] = J[i] = i+1;
      element_extraction(this, I, J, e.data(), e.desc());
      delete[] I;
      real_t acc_front = (hss_mat) ? hss_mat->check_compression(e.data(), e.desc()) : 0.0;
      if (!mpi_rank(this->front_comm)) {
	std::fstream fs("rank_stats_" + std::to_string(mpi_rank()) + ".log", std::fstream::out | std::fstream::app);
	fs << "DISTR \t " << this->sep << " \t " << this->dim_sep << " \t " << this->dim_upd
	   << " \t " << Rrow.cols() << " \t " << ((hss_mat) ? hss_mat->max_rank : -1)
	   << "\tapp err = " << acc_front << "\tcumm err = " << acc_total
	   << "\t[" << this->sep_begin << "," << this->sep_end << "], ";
	if (this->lchild && this->rchild)
	  fs << "lch=[" << this->lchild->sep_begin << "," << this->lchild->sep_end << "]/" << this->lchild->type()
	     << " rch=[" << this->rchild->sep_begin << "," << this->rchild->sep_end << "]/" << this->rchild->type();
	fs << std::endl;
	fs.close();
      }
    }

    if (this->lchild) this->lchild->release_work_memory();
    if (this->rchild) this->rchild->release_work_memory();

    if (this->dim_sep) {
      if (this->dim_upd) {
	MPI_Pcontrol(1, "DenseMPI::partially_factor");
	TIMER_TIME(HSS_PARTIALLY_FACTOR, 0, t_pfact);
	hss_mat->partially_factor(NULL, NULL, this->dim_sep);
	TIMER_STOP(t_pfact);
	MPI_Pcontrol(-1, "DenseMPI::partially_factor");
	MPI_Pcontrol(1, "DenseMPI::compute_schur");
	TIMER_TIME(HSS_COMPUTE_SCHUR, 0, t_comp_schur);
	hss_mat->compute_schur();
	TIMER_STOP(t_comp_schur);
	MPI_Pcontrol(-1, "DenseMPI::compute_schur");
      } else {
	MPI_Pcontrol(1, "DenseMPI::factor");
	TIMER_TIME(HSS_FACTOR, 0, t_fact);
	hss_mat->factor(NULL, NULL);
	TIMER_STOP(t_fact);
	MPI_Pcontrol(-1, "DenseMPI::factor");
      }
    }
    STRUMPACK_FLOPS(static_cast<long long>(hss_mat->flops_compr + hss_mat->flops_fact));
    //if (!etree_level && hss_mat) hss_mat->print_statistics();
    //hss_mat->print_statistics();
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::forward_multifrontal_solve
  (scalar_t* b_loc, DistributedMatrix<scalar_t,real_t>* b_dist, scalar_t* wmem, int etree_level, int task_depth) {
    if (this->visit(this->lchild)) this->lchild->forward_multifrontal_solve(b_loc, b_dist, wmem, etree_level+1, task_depth);
    if (this->visit(this->rchild)) this->rchild->forward_multifrontal_solve(b_loc, b_dist, wmem, etree_level+1, task_depth);
    DistributedMatrixWrapper<scalar_t,real_t> Bupd(this->ctxt, this->dim_upd, 1, wmem+this->p_wmem);
    Bupd.zero();
    DistributedMatrix<scalar_t,real_t> B(this->ctxt, this->dim_blk, 1);
    DistributedMatrix<scalar_t,real_t>& Bsep = b_dist[this->sep];
    this->look_left(Bsep, wmem);
    // B = [ Bsep; Bupd=wmem+p_wmem]
    pgemr2d(this->dim_sep, 1, Bsep.data(), 1, 1, Bsep.desc(), B.data(), 1, 1, B.desc(), this->ctxt_all);
    pgemr2d(this->dim_upd, 1, Bupd.data(), 1, 1, Bupd.desc(), B.data(), this->dim_sep+1, 1, B.desc(), this->ctxt_all);
    DistributedMatrix<scalar_t,real_t> X(this->ctxt, this->dim_blk, 1);
    if (this->dim_sep && this->dim_upd) {
      MPI_Pcontrol(1, "DenseMPI::reduce_RHS");
      TIMER_TIME(SOLVE_LOWER, 0, t_reduce);
      hss_mat->reduce_RHS(X.data(), X.desc(), B.data(), B.desc());
      TIMER_STOP(t_reduce);
      MPI_Pcontrol(-1, "DenseMPI::reduce_RHS");
    } else if (this->dim_sep) {// full ULV solve
      MPI_Pcontrol(1, "DenseMPI::solve");
      TIMER_TIME(SOLVE_LOWER_ROOT, 0, t_solve);
      hss_mat->solve(X.data(), X.desc(), B.data(), B.desc());
      TIMER_STOP(t_solve);
      MPI_Pcontrol(-1, "DenseMPI::solve");
    }
    STRUMPACK_FLOPS(static_cast<long long>(hss_mat->flops_solve));
    // result: X = [ Xsep; Xupd]
    pgemr2d(this->dim_sep, 1, X.data(), 1, 1, X.desc(), Bsep.data(), 1, 1, Bsep.desc(), this->ctxt_all);
    pgemr2d(this->dim_upd, 1, X.data(), this->dim_sep+1, 1, X.desc(), Bupd.data(), 1, 1, Bupd.desc(), this->ctxt_all);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::backward_multifrontal_solve
  (scalar_t* y_loc, DistributedMatrix<scalar_t,real_t>* y_dist, scalar_t* wmem, int etree_level, int task_depth) {
    DistributedMatrix<scalar_t,real_t>& Ysep = y_dist[this->sep];
    if (this->dim_upd) {
      if (this->dim_sep) {
	DistributedMatrixWrapper<scalar_t,real_t> Yupd(this->ctxt, this->dim_upd, 1, wmem+this->p_wmem);
	DistributedMatrix<scalar_t,real_t> Y(this->ctxt, this->dim_blk, 1);
	DistributedMatrix<scalar_t,real_t> X(this->ctxt, this->dim_blk, 1);
	// Y = [ Ysep; ...]
	// X = [ ...; wmem+p_wmem]
	pgemr2d(this->dim_sep, 1, Ysep.data(), 1, 1, Ysep.desc(), Y.data(), 1, 1, Y.desc(), this->ctxt_all);
	pgemr2d(this->dim_upd, 1, Yupd.data(), 1, 1, Yupd.desc(), X.data(), this->dim_sep+1, 1, X.desc(), this->ctxt_all);
	MPI_Pcontrol(1, "DenseMPI::expand_solution");
	TIMER_TIME(SOLVE_UPPER, 0, t_expand);
	hss_mat->expand_solution(X.data(), Y.desc(), Y.data(), Y.desc());
	TIMER_STOP(t_expand);
	MPI_Pcontrol(-1, "DenseMPI::expand_solution");
	// result: X = [ Xsep; Xupd]
	pgemr2d(this->dim_sep, 1, X.data(), 1, 1, X.desc(), Ysep.data(), 1, 1, Ysep.desc(), this->ctxt_all);
	pgemr2d(this->dim_upd, 1, X.data(), this->dim_sep+1, 1, X.desc(), Yupd.data(), 1, 1, Yupd.desc(), this->ctxt_all);
      }
    } // else do nothing, ULV solve already done in forward solve
    this->look_right(Ysep, wmem);
    if (this->visit(this->lchild)) this->lchild->backward_multifrontal_solve(y_loc, y_dist, wmem, etree_level+1, task_depth);
    if (this->visit(this->rchild)) this->rchild->backward_multifrontal_solve(y_loc, y_dist, wmem, etree_level+1, task_depth);
  }

  /**
   * Extract from (HSS - theta Vhat^* phi).
   *
   * Note that mat has the same context as this front, otherwise the
   * communication pattern would be hard to figure out.
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::get_submatrix_2d
  (integer_t* rows, integer_t* cols, DistributedMatrix<scalar_t,real_t>& mat, int task_depth) {
    if (this->front_comm == MPI_COMM_NULL || !this->dim_upd) return;
    std::vector<int> my_cols, orig_cols, my_rows, orig_rows;
    this->find_my_upd_indices(mat.cols(), cols, my_cols, orig_cols);
    this->find_my_upd_indices(mat.rows(), rows, my_rows, orig_rows);
    for (auto& i : my_rows) i++;
    for (auto& i : my_cols) i++;

    DistributedMatrix<scalar_t,real_t> e(this->ctxt, my_rows.size(), my_cols.size());
    if (my_rows.size() && my_cols.size()) {
      if (this->dim_sep) {
	MPI_Pcontrol(1, "DenseMPI::extract_schur");
	TIMER_TIME(HSS_EXTRACT_SCHUR, 3, t_ex_schur);
	hss_mat->extract_schur(e.data(), e.desc(), my_rows.data(), my_rows.size(), my_cols.data(), my_cols.size());
	TIMER_STOP(t_ex_schur);
	MPI_Pcontrol(-1, "DenseMPI::extract_schur");
      } else {
	int descA[9];
	descset(descA, this->dim_blk, this->dim_blk, Rrow.MB(), Rrow.NB(), 0, 0, this->blacs_context(), this->dim_blk);
	MPI_Pcontrol(1, "DenseMPI::extract");
	TIMER_TIME(HSS_EXTRACT_FULL, 3, t_ex_full);
	hss_mat->extract(NULL, descA, e.data(), e.desc(), my_rows.data(), my_rows.size(), my_cols.data(), my_cols.size());
	TIMER_STOP(t_ex_full);
	MPI_Pcontrol(-1, "DenseMPI::extract");
      }
    }
    auto P = mpi_nprocs(this->front_comm);
    std::vector<std::vector<scalar_t>> sbuf(P);
    e.extend_copy_to_buffers(orig_rows, orig_cols, mat, sbuf);

    scalar_t* rbuf=NULL, **pbuf=NULL;
    all_to_all_v(sbuf, rbuf, pbuf, this->front_comm);

    mat.extend_copy_from_buffers(orig_rows, orig_cols, e, pbuf);
    delete[] rbuf;
    delete[] pbuf;
  }

  template<typename scalar_t,typename real_t,typename integer_t> long long
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::get_node_factor_nonzeros() {
    return (hss_mat) ? static_cast<long long>(hss_mat->mem_ULV / sizeof(scalar_t)) : 0;
  }

  template<typename scalar_t,typename real_t,typename integer_t> integer_t
  FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>::get_maximum_rank(int task_depth) const {
    integer_t max_rank = (hss_mat) ? hss_mat->max_rank : 0;
    integer_t max_rank_l = 0;
    integer_t max_rank_r = 0;
    if (this->lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
      max_rank_l = this->lchild->get_maximum_rank(task_depth+1);
    if (this->rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
      max_rank_r = this->rchild->get_maximum_rank(task_depth+1);
#pragma omp taskwait
    return std::max(max_rank, std::max(max_rank_l, max_rank_r));
  }

} // end namespace strumpack

#endif //FRONTAL_MATRIX_HSS_MPI_HPP
