/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef HSSTOOLS_H
#define HSSTOOLS_H

#include <cmath>
#include <complex>
#include <algorithm>
#include <iterator>
#include <vector>
#include "blas_lapack_wrapper.hpp"
#include "blas_lapack_omp_task.hpp"
#include "strumpack_parameters.hpp"

namespace strumpack {

  /**
   * Modified Gram-Schmidt with column pivoting (RRQR), followed by triangular solve.
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  column_pivoted_modified_Gram_Schmidt_ID
  (integer_t m, integer_t n, scalar_t* a, integer_t lda, scalar_t* r, integer_t ldr,
   int* piv, int* ind, integer_t& rank, real_t rtol, real_t atol, int depth) {
    bool tasked = depth<params::task_recursion_cutoff_level;
    real_t* col_norms = new real_t[n];
    if (tasked) {
      integer_t loop_tasks = std::max(params::num_threads / (depth+1), 1);
      integer_t B = std::max(n / loop_tasks, integer_t(1));
      for (integer_t task=0; task<std::ceil(n/float(B)); task++) {
#pragma omp task default(shared) firstprivate(task)
	{
	  for (integer_t i=task*B; i<std::min((task+1)*B,n); i++) {
	    col_norms[i] = blas::nrm2<scalar_t,real_t>(m, &a[i*lda], 1);
	    col_norms[i] *= col_norms[i];
	    ind[i] = piv[i] = i;
	  }
	}
      }
#pragma omp taskwait
    } else {
      for (integer_t i=0; i<n; i++) {
	col_norms[i] = blas::nrm2<scalar_t,real_t>(m, &a[i*lda], 1);
	col_norms[i] *= col_norms[i];
	ind[i] = piv[i] = i;
      }
    }

    real_t r00 = real_t(1.);
    integer_t i;
    for (i=0; i<std::min(m,n); i++) {
      integer_t max_col = std::distance(col_norms, std::max_element(&col_norms[i], &col_norms[n]));
      // recompute norm (should be more accurate than using updated norms)
      real_t rii = blas::nrm2<scalar_t,real_t>(m, &a[max_col*lda], 1);
      if (i==0) { r00 = rii; if (r00<=atol) break; }
      else if (rii/r00<=rtol || rii<=atol) break;
      if (max_col != i) {
	piv[i] = max_col;
	std::swap(ind[i], ind[max_col]);
	std::swap(col_norms[i], col_norms[max_col]);
	blas::swap(m, &a[i*lda], 1, &a[max_col*lda], 1);
	blas::swap(i, &r[i*ldr], 1, &r[max_col*ldr], 1);
      }
      r[i*ldr+i] = rii;
      blas::scal(m, scalar_t(1.)/rii, &a[i*lda], 1);  // normalize vector i
      if (i < n-1) {  // make columns i+1 to n orthogonal to column i
	if (tasked) {
	  gemv_omp_task('C', m, n-(i+1), scalar_t(1.), &a[(i+1)*lda], lda, &a[i*lda], 1, scalar_t(0.), &r[(i+1)*ldr+i], ldr, depth);
	  blas::lacgv(n-(i+1), &r[(i+1)*ldr+i], ldr);
	  geru_omp_task(m, n-(i+1), scalar_t(-1), &a[i*lda], 1, &r[(i+1)*ldr+i], ldr, &a[(i+1)*lda], lda, depth);
	} else {
	  blas::gemv('C', m, n-(i+1), scalar_t(1.), &a[(i+1)*lda], lda, &a[i*lda], 1, scalar_t(0.), &r[(i+1)*ldr+i], ldr);
	  blas::lacgv(n-(i+1), &r[(i+1)*ldr+i], ldr);
	  blas::geru(m, n-(i+1), scalar_t(-1), &a[i*lda], 1, &r[(i+1)*ldr+i], ldr, &a[(i+1)*lda], lda);
	}
	for (integer_t j=i+1; j<n; j++) col_norms[j] -= std::real(std::conj(r[j*ldr+i])*r[j*ldr+i]);
      }
    }
    rank = i;

    // triangular solve, introduces identity block
    if (tasked) trsm_omp_task('L', 'U', 'N', 'N', rank, n-rank, scalar_t(1.), r, ldr, &r[rank*ldr], ldr, depth);
    else blas::trsm('L', 'U', 'N', 'N', rank, n-rank, scalar_t(1.), r, ldr, &r[rank*ldr], ldr);
    // explicitly set identity matrix
    blas::laset('F', rank, rank, scalar_t(0.), scalar_t(1.), r, ldr);

    delete[] col_norms;
  }


  template<typename scalar_t,typename real_t,typename integer_t> void
  column_pivoted_modified_Gram_Schmidt_ID_BLAS3
  (int m, int n, scalar_t* a, int lda, scalar_t* r, int ldr,
   int* piv, int* ind, integer_t& rank, real_t rtol, real_t atol, int depth) {
    real_t* col_norms = new real_t[n];
    for (int i=0; i<n; i++) {
      col_norms[i] = blas::nrm2<scalar_t,real_t>(m, a+i*lda, 1);
      col_norms[i] *= col_norms[i];
      ind[i] = piv[i] = i;
    }

    real_t r00 = real_t(1.);
    int i = 0, j = 0;
    int NB = 32;
    int min_mn = std::min(m,n);
    bool rank_revealed = false;
    while (i < min_mn && !rank_revealed) {
      for (i=j; i<std::min(min_mn,j+NB); i++) {
	int max_col = std::distance(col_norms, std::max_element(col_norms+i, col_norms+n));
	if (max_col != i) {
	  piv[i] = max_col;
	  std::swap(ind[i], ind[max_col]);
	  std::swap(col_norms[i], col_norms[max_col]);
	  blas::swap(m, a+i*lda, 1, a+max_col*lda, 1);
	  blas::swap(i, r+i*ldr, 1, r+max_col*ldr, 1);
	}
	// update the pivot column with what we have until now in this block, cfr geru
	blas::gemv('N', m, i-j, scalar_t(-1.), a+j*lda, lda, r+i*ldr+j, 1, scalar_t(1.), a+i*lda, 1);
	// recompute norm (should be more accurate than using updated norms)
	real_t rii = blas::nrm2<scalar_t,real_t>(m, a+i*lda, 1);
	if (i==0) { r00 = rii; if (r00<=atol) { rank_revealed = true; break; } }
	else if (rii/r00<=rtol || rii<=atol) { rank_revealed = true; break; }
	r[i*ldr+i] = rii;
	blas::scal(m, scalar_t(1.)/rii, a+i*lda, 1);  // normalize vector i
	if (i < n-1) {  // make columns i+1 to n orthogonal to column i
	  blas::gemv('C', m, n-(i+1), scalar_t(1.), a+(i+1)*lda, lda, a+i*lda, 1, scalar_t(0.), r+(i+1)*ldr+i, ldr);
	  blas::lacgv(n-(i+1), r+(i+1)*ldr+i, ldr);
	  // this rank-1 update is delayed until all NB columns are handled, then it becomes a gemm
	  //blas::geru(m, n-(i+1), scalar_t(-1), a+i*lda, 1, r+(i+1)*ldr+i, ldr, a+(i+1)*lda, lda);
	  for (int k=i+1; k<n; k++) col_norms[k] -= std::real(std::conj(r[k*ldr+i])*r[k*ldr+i]);
	}
      }
      if (i < n-1) // this single gemm replaces the NB geru's
	blas::gemm('N', 'N', m, n-(i+1), i-j, scalar_t(-1.), a+j*lda, lda, r+j+(i+1)*ldr, ldr,
		   scalar_t(1.), a+(i+1)*lda, lda);
      j += NB;
    }
    delete[] col_norms;
    rank = i;

    // triangular solve, introduces identity block
    blas::trsm('L', 'U', 'N', 'N', rank, n-rank, scalar_t(1.), r, ldr, r+rank*ldr, ldr);
    // explicitly set identity matrix
    blas::laset('F', rank, rank, scalar_t(0.), scalar_t(1.), r, ldr);
  }


  template<typename scalar_t,typename real_t,typename integer_t> void
  BLAS3_ID(integer_t m, integer_t n, scalar_t* a, integer_t lda, int* piv,
	   int* ind, integer_t& rank, real_t rtol, real_t atol, int depth) {
    int info;
    scalar_t* tau = new scalar_t[std::max(integer_t(1),std::min(m, n))];
    for (int i=0; i<n; i++) ind[i]=0;
    int int_rank;
    blas::geqp3tol<scalar_t,real_t>(m, n, a, lda, ind, tau, &info, int_rank, rtol, atol, depth);
    rank = integer_t(int_rank);
    // Generate P from J
    for (int i=1; i<=n; i++) {
      int j = ind[i-1];
      while (j<i) j = ind[j-1];
      piv[i-1] = j;
    }
    for (int i=0; i<n; i++) {
      piv[i]--;
      ind[i]--;
    }
    bool tasked = depth<params::task_recursion_cutoff_level;
    // triangular solve, introduces identity block
    if (tasked) trsm_omp_task('L', 'U', 'N', 'N', rank, n-rank, scalar_t(1.), a, lda, &a[rank*lda], lda, depth);
    else blas::trsm('L', 'U', 'N', 'N', rank, n-rank, scalar_t(1.), a, lda, &a[rank*lda], lda);
    // explicitly set identity matrix
    blas::laset('F', rank, rank, scalar_t(0.), scalar_t(1.), a, lda);
    delete[] tau;
  }

  // if succesful, UV is allocated and filled with the basis matrix,
  // else UV is NULL
  template<typename scalar_t,typename real_t,typename integer_t> void
  ID(integer_t rows, integer_t cols, scalar_t* S, integer_t lds, scalar_t*& UV, scalar_t* work,
     int* piv, int* ind, integer_t& rank, HSSStrategy<real_t>& strat, int depth) {
#if defined(CPMGS_ID)
    scalar_t* temp = new scalar_t[rows*cols];
    blas::omatcopy('C', rows, cols, S, lds, temp, cols);
    column_pivoted_modified_Gram_Schmidt_ID_BLAS3<scalar_t,real_t,integer_t>
      (cols, rows, temp, cols, work, cols, piv, ind, rank, strat.rctol, strat.actol, depth);
    delete[] temp;
#else
    blas::omatcopy('C', rows, cols, S, lds, work, cols);
    BLAS3_ID<scalar_t,real_t,integer_t>(cols, rows, work, cols, piv, ind, rank, strat.rctol, strat.actol, depth);
#endif
    if (UV) delete[] UV;
    UV = NULL;
    if ((strat.rank_pattern == params::ADAPTIVE) &&
	(cols - rank < params::rank_offset) &&
	(cols < params::maximum_rank))
      return;
    else {
      UV = new scalar_t[rows*rank]; // TODO do not store identity part!!!!
      blas::omatcopy('C', rank, rows, work, cols, UV, rows);
    }
  }

} // end namespace strumpack

#endif // HSSTOOLS_H
